from gettext import gettext as _
import re

from click.shell_completion import add_completion_class, BashComplete


# inspired by https://github.com/pallets-eco/click-bash4.2-completion
_bash_42_source = """\
%(complete_func)s() {
    local IFS=$'\\n'
    local response
    response=$(env COMP_WORDS="${COMP_WORDS[*]}" COMP_CWORD=$COMP_CWORD \
%(complete_var)s=bash_complete $1)
    for completion in $response; do
        IFS=',' read type value <<< "$completion"
        if [[ $type == 'dir' ]]; then
            COMPREPLY=()
            compopt -o dirnames
        elif [[ $type == 'file' ]]; then
            COMPREPLY=()
            compopt -o default
        elif [[ $type == 'plain' ]]; then
            COMPREPLY+=($value)
        fi
    done
    return 0
}
%(complete_func)s_setup() {
    local COMPLETION_OPTIONS=""
    local BASH_VERSION_ARR=(${BASH_VERSION//./ })
    # Only BASH version 4.4 and later have the nosort option.
    if [ ${BASH_VERSION_ARR[0]} -gt 4 ] || ([ ${BASH_VERSION_ARR[0]} -eq 4 ] \
&& [ ${BASH_VERSION_ARR[1]} -ge 4 ]); then
        COMPLETION_OPTIONS="-o nosort"
    fi
    complete $COMPLETION_OPTIONS -F %(complete_func)s %(prog_name)s
}
%(complete_func)s_setup;
"""


@add_completion_class
class _PatchedBashComplete(BashComplete):
    """ Patched Shell completion for Bash """
    source_template = _bash_42_source
    name = 'bash'  # this will override the default bash provided by click

    # change the original version check
    def _check_version(self) -> None:
        import subprocess

        output = subprocess.run(
            ["bash", "-c", "echo ${BASH_VERSION}"], stdout=subprocess.PIPE, check=False
        )
        match = re.search(r"^(\d+)\.(\d+)\.\d+", output.stdout.decode())

        if match is not None:
            major, minor = match.groups()

            if major < "4" or major == "4" and minor < "2":
                raise RuntimeError(
                    _(
                        "Shell completion is not supported for Bash"
                        " versions older than 4.2."
                    )
                )
        else:
            raise RuntimeError(
                _("Couldn't detect Bash version, shell completion is not supported.")
            )



import base64
import os
from pathlib import Path
import shutil
import yaml
from .config import ConfigManager
from ..britive_cli import BritiveCli


def sanitize(name: str):
    name = name.lower()
    # name = name.replace(' ', '_').replace('/', "_").replace('\\', '_')
    return name


def check_env_var(filename, cli: BritiveCli):
    kubeconfig = os.getenv('KUBECONFIG')

    # no env var present
    if not kubeconfig:
        command = f'export KUBECONFIG=~/.kube/config:{filename}'
        cli.print('Please ensure your KUBECONFIG environment variable includes the Britive managed kube config file.')
        cli.print(command)
    else:
        for configfile in kubeconfig.split(':'):
            full_path = str(Path(configfile).expanduser()).lower()
            if filename.lower() == full_path:
                return  # we found what we came for - silently continue

        # if we get here we need to instruct the user to add the britive managed kube config file
        cli.print('Please modify your KUBECONFIG environment variable to include the Britive managed kube config file.')
        command = f'export KUBECONFIG="${{KUBECONFIG}}:{filename}"'
        cli.print(command)


def merge_new_with_existing(clusters, contexts, users, filename, tenant):
    # get the existing config, so we can pop out all
    # items related to this tenant as we will be replacing
    # them with the above created items
    existing_kubeconfig = {}
    if Path(filename).exists():
        with open(filename, 'r', encoding='utf-8') as f:
            existing_kubeconfig = yaml.safe_load(f) or {}

    prefix = f'{tenant}-'
    for cluster in existing_kubeconfig.get('clusters', []):
        if not cluster.get('name', '').startswith(prefix):
            clusters.append(cluster)

    for context in existing_kubeconfig.get('contexts', []):
        cluster_name = context.get('context', {}).get('cluster', '')
        if not cluster_name.startswith(prefix):
            contexts.append(context)

    for user in existing_kubeconfig.get('users', []):
        if not user.get('name', '').startswith(prefix):
            users.append(user)

    kubeconfig = {'apiVersion': 'v1', 'clusters': clusters, 'contexts': contexts, 'users': users, 'kind': 'Config'}

    # write out the config file
    with open(filename, 'w', encoding='utf-8') as f:
        yaml.safe_dump(kubeconfig, f, default_flow_style=False, encoding='utf-8')


def parse_profiles(profiles, aliases):
    cluster_names = {}
    assigned_aliases = []
    for profile in profiles:
        env_profile = f"{sanitize(profile['env'])}-{sanitize(profile['profile'].lower())}"
        if env_profile not in cluster_names:
            app = BritiveCli.escape_profile_element(profile['app'])
            env = BritiveCli.escape_profile_element(profile['env'])
            pro = BritiveCli.escape_profile_element(profile['profile'])

            escaped_profile_str = f"{app}/{env}/{pro}".lower()
            alias = aliases.get(escaped_profile_str, None)
            assigned_aliases.append(alias)

            cluster_names[env_profile] = {
                'apps': [],
                'url': profile['url'],
                'cert': profile['cert'],
                'escaped_profile': escaped_profile_str,
                'profile': f"{profile['app']}/{profile['env']}/{profile['profile']}".lower(),
                'alias': alias,
            }
        cluster_names[env_profile]['apps'].append(sanitize(profile['app']))
    return [cluster_names, assigned_aliases]


def valid_cert(cert: str, profile: str, cli: BritiveCli):
    try:
        decoded_cert = base64.b64decode(cert).decode('utf-8')
        if not decoded_cert.startswith('-----BEGIN CERTIFICATE-----'):
            raise ValueError()
        if not decoded_cert.strip().endswith('-----END CERTIFICATE-----'):
            raise ValueError()
        return True
    except Exception:
        cli.print(f'could not properly decode certificate authority data for profile {profile} - skipping this cluster')
        return False


def build_tenant_config(tenant, cluster_names, username, cli: BritiveCli):
    kube_exec_full_path = shutil.which('pybritive-kube-exec')
    if not kube_exec_full_path:
        kube_exec_full_path = 'pybritive-kube-exec'
    users = (
        [
            {
                'name': username,
                'user': {
                    'exec': {
                        'apiVersion': 'client.authentication.k8s.io/v1beta1',
                        'command': kube_exec_full_path,
                        'args': ['-t', tenant],
                        'env': None,
                        'interactiveMode': 'Never',
                        'provideClusterInfo': True,
                    }
                },
            }
        ]
        if len(cluster_names.keys()) > 0
        else []
    )
    contexts = []
    clusters = []

    for env_profile, details in cluster_names.items():
        if len(details['apps']) == 1:
            names = [env_profile]
        else:
            names = [f"{sanitize(a)}-{env_profile}" for a in details['apps']]

        cert = details['cert']
        url = details['url']

        if not valid_cert(cert=cert, profile=details['profile'], cli=cli):
            continue

        for name in names:
            clusters.append(
                {
                    'name': f'{tenant}-{name}',
                    'cluster': {
                        'certificate-authority-data': cert,
                        'server': url,
                        'extensions': [
                            {
                                'name': 'client.authentication.k8s.io/exec',
                                'extension': {'britive-profile': details.get('alias') or details['escaped_profile']},
                            }
                        ],
                    },
                }
            )

            contexts.append(
                {
                    'name': details.get('alias') or f'{tenant}-{name}',
                    'context': {'cluster': f'{tenant}-{name}', 'user': username},
                }
            )
    return [clusters, contexts, users]


def build_kube_config(profiles: list, config: ConfigManager, username: str, cli: BritiveCli):
    tenant = config.get_tenant()['alias'].lower()  # must be run first to set the tenant alias in the config

    # something unique that is not likely to clash with any other username that may be present in a kube config file
    # add the tenant details which will mean 1 user per tenant
    username = f'{tenant}-{username}'

    # grab the aliases
    aliases = config.get_profile_aliases(reverse_keys=True)

    # parse all the profiles
    cluster_names, assigned_aliases = parse_profiles(profiles, aliases)

    # establish the 3 elements of the config
    clusters, contexts, users = build_tenant_config(
        tenant=tenant, cluster_names=cluster_names, username=username, cli=cli
    )

    # calculate the path for the config
    kube_dir = Path(config.base_path) / 'kube'
    kube_dir.mkdir(exist_ok=True)
    filename = str(kube_dir / 'config')

    # merge any existing config with the new config
    # and write it to disk
    merge_new_with_existing(clusters=clusters, contexts=contexts, users=users, tenant=tenant, filename=filename)

    # if required ensure we tell the user they need to modify their KUBECONFIG env var
    # in order to pick up the Britive managed kube config file
    if len(clusters) > 0:
        check_env_var(filename=filename, cli=cli)



import platform

import click
import pkg_resources


# for --version we need to print the version and exit - nothing more, nothing less
def version_callback(ctx, self, value):
    if value:
        cli_version = pkg_resources.get_distribution('pybritive').version
        click.echo(f'pybritive: {cli_version} / platform: {platform.platform()} / python: {platform.python_version()}')
        raise click.exceptions.Exit()


option = click.option(
    '--version', '-v',
    default=None,
    callback=version_callback,
    is_eager=True,
    help='Prints the PyBritive CLI version and exits.',
    is_flag=True
)



def common_asserts(result):
    assert result.exit_code == 0
    assert 'pybritive:' in result.output
    assert '/ platform:' in result.output
    assert '/ python: 3.' in result.output


def test_version_full_flag(runner, cli):
    with runner.isolated_filesystem():
        result = runner.invoke(cli, ['--version'])
        common_asserts(result)


def test_version_short_flag(runner, cli):
    with runner.isolated_filesystem():
        result = runner.invoke(cli, ['-v'])
        common_asserts(result)



from dataclasses import dataclass
from functools import wraps

import click
from merge_args import merge_args
from ..britive_cli import BritiveCli
from click import Context


@dataclass
class Common:
    britive: BritiveCli


def should_set_output_format(ctx: Context) -> bool:
    parent_command = ctx.parent.command.name
    command = ctx.command.name
    if parent_command in ['configure', 'clear']:
        return False

    if parent_command == 'cache' and command in ['clear']:
        return False

    return True


# this wrapper exists to centralize all "common" CLI options (options that exist for all commands)
def build_britive(f):
    @merge_args(f)
    @wraps(f)
    @click.pass_context
    def wrapper(
            ctx,
            **kwargs
    ):
        ctx.obj = Common(BritiveCli(
            tenant_name=kwargs.get('tenant'),
            token=kwargs.get('token'),
            silent=kwargs.get('silent', False),
            passphrase=kwargs.get('passphrase'),
            federation_provider=kwargs.get('federation_provider')
        ))
        if should_set_output_format(ctx=ctx):
            ctx.obj.britive.set_output_format(kwargs.get('output_format'))
        return f(ctx=ctx, **kwargs)
    return wrapper



from ..helpers.cache import Cache
from ..helpers.config import ConfigManager


def profile_completer(ctx, param, incomplete):
    profiles = Cache().get_profiles()
    try:
        config = ConfigManager(None)
        config.load()
        aliases = config.profile_aliases
        alias_list = []
        if aliases:
            alias_list = list(aliases.keys())
    except Exception:
        alias_list = []

    options = profiles + alias_list

    return [o for o in options if incomplete.lower() in o.lower()]



import inspect
import typing as t

from britive.britive import Britive
import pkg_resources


def get_dynamic_method_parameters(method):
    try:
        # create an instance of the Britive class, so we can inspect it
        # this doesn't need to actually connect to any tenant, and we couldn't even if we
        # wanted to since when performing shell completion we have no tenant/token
        # context in order to properly establish a connection.
        b = Britive(token='ignore', tenant='britive.com', query_features=False)

        # parse the method, so we can determine where in the "hierarchy" we are
        # and what commands/subcommands the user should be presented with

        for part in method.split('.'):
            b = getattr(b, part)

        params = {}
        spec = inspect.getfullargspec(b)
        # reformat parameters into a more consumable dict while holds all the required details
        helper = spec[6]
        helper.pop('return', None)

        for param, param_type in helper.items():
            params[param] = {}

        defaults = [] if spec[3] is None else list(spec[3])
        names = [] if spec[0] is None else list(spec[0])

        if len(defaults) > 0:
            for i in range(1, len(defaults) + 1):
                name = names[-1 * i]
                default = defaults[-1 * i]
                params[name]['default'] = '<empty string>' if default == '' else default

        try:  # we don't REALLY need the doc string so if there are errors just eat them and move on
            doc_lines = inspect.getdoc(b)
            doc_lines = doc_lines.replace(':returns:', 'RETURNSPLIT')
            doc_lines = doc_lines.replace(':return:', 'RETURNSPLIT')
            doc_lines = doc_lines.split('RETURNSPLIT', maxsplit=1)[0].split(':param ')[1:]

            for line in doc_lines:
                helper = line.split(':')
                name = helper[0].strip()
                help_text = ''.join(helper[1].strip().splitlines()).replace('    ', ' ')
                params[name]['help'] = help_text
        except:
            pass

        param_list = []

        for name, values in params.items():
            help_text = values.get('help') or ''

            if 'default' in values.keys():  # cannot do a .get('default') as the default value could be False/None/etc.
                preamble = f'[optional: default = {values["default"]}]'
                if help_text == '':
                    help_text = preamble
                else:
                    help_text = f'{preamble} - {help_text}'

            param = {
                'flag': f'--{name.replace("_", "-")}',
                'help': help_text
            }

            param_list.append(param)

        param_list.append({
            'flag': '---------------------',
            'help': 'separator between sdk parameters and cli parameters'
        })

        return param_list
    except Exception:
        return []


def command_api_patch_shell_complete(cls):
    # click < 8.0.0 does shell completion different...
    # not all the classes/decorators are available, so we cannot
    # create custom shell completions like we can with click > 8.0.0
    major, minor = pkg_resources.get_distribution('click').version.split('.')[:2]

    # we cannot patch the shell_complete method because it does not exist (click 7.x doesn't have it)
    # future proofing this as well in case click 9.x changes things up a lot
    if int(major) != 8:
        return

    # we could potentially patch but there could be changes to shell_complete method which are not
    # accounted for in this patch - we will have to manually review any changes and ensure they are
    # backwards compatible.
    if int(minor) != 1:
        return

    from click.shell_completion import CompletionItem
    from click.core import ParameterSource
    from click import Context, Option

    # https://stackoverflow.com/questions/43778914/python3-using-super-in-eq-methods-raises-runtimeerror-super-class
    __class__ = cls  # provide closure cell for super()

    def shell_complete(self, ctx: Context, incomplete: str) -> t.List["CompletionItem"]:
        from click.shell_completion import CompletionItem  # here since this method will be monkey patched in

        results: t.List["CompletionItem"] = []

        if incomplete and not incomplete[0].isalnum():
            method = ctx.params.get('method')
            if method:
                dynamic_params = get_dynamic_method_parameters(method)

                results.extend(
                    CompletionItem(p['flag'], help=p['help'])
                    for p in dynamic_params if p['flag'].startswith(incomplete)
                )

            for param in self.get_params(ctx):
                if (
                    not isinstance(param, Option)
                    or param.hidden
                    or (
                        not param.multiple
                        and ctx.get_parameter_source(param.name)  # type: ignore
                        is ParameterSource.COMMANDLINE
                    )
                ):
                    continue

                results.extend(
                    CompletionItem(name, help=param.help)
                    for name in [*param.opts, *param.secondary_opts]
                    if name.startswith(incomplete)
                )

        results.extend(super().shell_complete(ctx, incomplete))

        return results

    cls.shell_complete = shell_complete



import os
import sys
import click
from .options.britive_options import britive_options
from .commands.user import user as command_user
from .commands.configure import configure as group_configure
from .commands.login import login as command_login
from .commands.logout import logout as command_logout
from .commands.ls import ls as group_ls
from .commands.checkout import checkout as command_checkout
from .commands.checkin import checkin as command_checkin
from .commands.secret import secret as group_secret
from .commands.cache import cache as group_cache
from .commands.request import request as group_request
from .commands.clear import clear as group_clear
from .commands.api import api as command_api
from .commands.ssh import ssh as group_ssh
from .commands.aws import aws as group_aws


def safe_cli():
    debug = os.getenv('PYBRITIVE_DEBUG')
    try:
        if not debug:
            sys.tracebacklimit = 0
        cli()
    except Exception as e:
        if debug:
            raise e
        raise e from None


# this is the "main" app - it really does nothing but print the overview/help section
@click.group()
@britive_options(names='version')
def cli(version):
    """
    PyBritive CLI - Pure Python Implementation for a Britive CLI
    """


cli.add_command(command_user)
cli.add_command(group_ls)
cli.add_command(command_login)
cli.add_command(command_logout)
cli.add_command(command_checkout)
cli.add_command(group_configure)
cli.add_command(command_checkin)
cli.add_command(group_secret)
cli.add_command(group_cache)
cli.add_command(group_request)
cli.add_command(group_clear)
cli.add_command(command_api)
cli.add_command(group_ssh)
cli.add_command(group_aws)


if __name__ == "__main__":
    safe_cli()



def common_asserts(result, substring=None, exit_code=0):
    assert result.exit_code == exit_code
    if isinstance(substring, str):
        substring = [substring]
    for sub in substring:
        assert sub in result.output


def test_ls_profiles(runner, cli):
    result = runner.invoke(cli, 'ls profiles -f yaml'.split(' '))
    common_asserts(result, ['Application', 'Description', 'Environment', 'Profile', 'Type'])


def test_ls_applications(runner, cli):
    result = runner.invoke(cli, 'ls applications -f yaml'.split(' '))
    common_asserts(result, ['Application', 'Description', 'Type'])


def test_ls_environments(runner, cli):
    result = runner.invoke(cli, 'ls environments -f yaml'.split(' '))
    common_asserts(result, ['Application', 'Description', 'Environment', 'Type'])


def test_ls_secrets(runner, cli):
    result = runner.invoke(cli, 'ls secrets -f yaml'.split(' '))
    common_asserts(
        result,
        ['entityType', 'id', 'metadata', 'name', 'path', 'rotationInterval', 'secretNature', 'secretType', 'status'],
    )



import click
from ..helpers.build_britive import build_britive
from ..options.britive_options import britive_options


@click.group()
def cache():
    """Manage local cache settings."""
    pass


@cache.command()
@build_britive
@britive_options(names='tenant,token,silent,passphrase,federation_provider')
def profiles(ctx, tenant, token, silent, passphrase, federation_provider):
    """Cache profiles locally to facilitate auto-completion of profile names on checkin/checkout."""
    ctx.obj.britive.cache_profiles(from_cache_command=True)


@cache.command()
@build_britive
@britive_options(names='tenant,token,silent,passphrase,federation_provider')
def kubeconfig(ctx, tenant, token, silent, passphrase, federation_provider):
    """Cache a Britive managed kube config file based on the profiles to which the caller has access."""
    ctx.obj.britive.construct_kube_config(from_cache_command=True)


@cache.command()
@build_britive
def clear(ctx):
    """Clears the local cache."""
    ctx.obj.britive.cache_clear()



import click
import pkg_resources


click_major_version = int(pkg_resources.get_distribution('click').version.split('.')[0])


def click_smart_api_method_argument(func):
    if click_major_version >= 8:
        from ..completers.api import api_completer
        dec = click.argument('method', shell_complete=api_completer)
    else:
        dec = click.argument('method')
    return dec(func)



from pathlib import Path
import os


def test_logout(runner, cli):
    result = runner.invoke(cli, ['logout'])
    assert result.exit_code == 0
    local_home = os.getenv('PYBRITIVE_HOME_DIR')
    path = Path(Path(local_home) / '.britive' / 'pybritive.credentials.encrypted')
    with open(str(path), 'r', encoding='utf-8') as f:
        data = f.read()
    assert len(data) == 0



import click
from ..choices.profile_type import profile_type_choices


option = click.option(
    '--profile-type',
    type=profile_type_choices,
    show_choices=True,
    default=None,
    help='The profile type to use when checking out/in a profile.'
)



import click
from ..choices.output_format import output_format_choices
from ..choices.backend import backend_choices
from ..options.britive_options import britive_options
from ..helpers.build_britive import build_britive


@click.group()
def configure():
    """Configures the PyBritive CLI."""


@configure.command()
@build_britive
@britive_options(names='configure_tenant,configure_alias,format,configure_prompt')
def tenant(ctx, configure_tenant, configure_alias, output_format, configure_prompt):
    """Configures tenant level settings for the PyBritive CLI.

    If CLI options/flags are not provided an interactive data entry process will collect any needed data.
    """
    tenant_name = configure_tenant
    no_prompt = configure_prompt
    alias = configure_alias

    if not no_prompt:
        if not tenant_name:
            tenant_name = click.prompt(
                'The name of the tenant: [tenant].britive-app.com',
                type=str
            )
        if not alias:
            alias = click.prompt(
                'Optional alias for the above tenant. This alias would be used with the `--tenant` flag',
                default=tenant_name,
                type=str
            )
        if not output_format:
            output_format = click.prompt(
                'Output format (csv|json|table|yaml): ',
                default='json',
                type=output_format_choices
            )

    if not tenant_name or len(tenant_name.strip()) == 0:
        raise click.ClickException('Tenant Name not provided.')
    if not alias:
        alias = tenant_name
    if not output_format or output_format not in output_format_choices.choices:
        ctx.obj.britive.print(f'Invalid output format {output_format} provided. Defaulting to "json".')
        output_format = 'json'

    ctx.obj.britive.configure_tenant(
        tenant=tenant_name,
        alias=alias,
        output_format=output_format
    )


@configure.command(name='global')  # have to specify the name since global is a reserved word
@build_britive
@britive_options(names='configure_tenant,format,configure_prompt,configure_backend')
def global_command(ctx, configure_tenant, output_format, configure_prompt, configure_backend):
    """Configures global level settings for the PyBritive CLI.

    If CLI options/flags are not provided an interactive data entry process will collect any needed data.
    """
    default_tenant_name = configure_tenant
    no_prompt = configure_prompt
    backend = configure_backend

    if not no_prompt:
        if not default_tenant_name:
            default_tenant_name = click.prompt(
                'The default tenant name',
                type=str
            )
        if not output_format:
            output_format = click.prompt(
                'Output format (csv|json|table|yaml)',
                default='json',
                type=output_format_choices
            )
        if not output_format or output_format not in output_format_choices.choices:
            ctx.obj.britive.print(f'Invalid output format {output_format} provided. Defaulting to "json".')
            output_format = 'json'
        if not backend:
            backend = click.prompt(
                'Britive temporary credential storage backend',
                default='file',
                type=backend_choices
            )

    ctx.obj.britive.configure_global(
        default_tenant_name=default_tenant_name,
        output_format=output_format,
        backend=backend
    )


@configure.command(name='import')  # have to specify the name since import is a reserved word
@build_britive
@britive_options(names='silent,token,federation_provider')
def import_command(ctx, silent, token, federation_provider):  # silent is handled by @build_britive
    """Import an existing configuration from the Node.js/NPM version of the Britive CLI."""
    ctx.obj.britive.import_existing_npm_config()


@configure.command()
@build_britive
@britive_options(names='silent')
@click.argument('section')
@click.argument('field')
@click.argument('value')
def update(ctx, silent, section, field, value):  # silent is handled by @build_britive
    """Provides a mechanism to directly update any section/field/value in the config file.

    All arguments provided will be converted to lowercase before being persisted.

    `SECTION`: The config section (example: global, tenant-foo)

    `FIELD`: The field within the section (example: default_tenant, name, output_format)

    `VALUE`: The value of the field.

    Example: `pybritive configure update global output_format json`
    """
    section = section.lower().strip()
    field = field.lower().strip()
    value = value.lower().strip()
    ctx.obj.britive.configure_update(section=section, field=field, value=value)



import click
from ..helpers.build_britive import build_britive
from ..options.britive_options import britive_options


@click.command()
@build_britive
@britive_options(names='tenant,token,silent,passphrase,federation_provider,browser')
def login(ctx, tenant, token, silent, passphrase, federation_provider, browser):
    """Perform an interactive login to obtain temporary credentials.

    This only applies when an API token has not been specified via `--token,-T` or via environment variable
    `BRITIVE_API_TOKEN`.
    """
    ctx.obj.britive.login(explicit=True, browser=browser)



import hashlib
import json
import os
from pathlib import Path
import time
from .encryption import StringEncryption, InvalidPassphraseException


class Cache:
    def __init__(self, passphrase: str = None):
        self.passphrase = passphrase
        self.string_encryptor = StringEncryption(passphrase=self.passphrase)
        home = os.getenv('PYBRITIVE_HOME_DIR', str(Path.home()))
        self.base_path = str(Path(home) / '.britive')
        self.path = str(Path(self.base_path) / 'pybritive.cache')  # handle os specific separators properly
        self.cache = {}
        self.default_key_values = {
            'profiles': [],
            'awscredentialprocess': {},
            'kube-exec': {},
            'banners': {}
        }
        self.load()

    def load(self):
        path = Path(self.path)
        if not path.is_file():  # cache file does not yet exist, create it as an empty file
            path.parent.mkdir(exist_ok=True, parents=True)
            cache = {
                'profiles': []
            }
            path.write_text(json.dumps(cache, indent=2, default=str), encoding='utf-8')

        with open(str(self.path), 'r', encoding='utf-8') as f:
            try:
                self.cache = json.loads(f.read())
            except json.decoder.JSONDecodeError:
                self.cache = {}

        self.cache = {**self.default_key_values, **self.cache}

    def write(self):
        # write the new cache file
        with open(str(self.path), 'w', encoding='utf-8') as f:
            f.write(json.dumps(self.cache, indent=2, default=str))

    def get_profiles(self):
        return self.cache.get('profiles', [])

    def save_profiles(self, profiles: list):
        self.cache['profiles'] += profiles
        # dedup the list of profiles
        self.cache['profiles'] = list(dict.fromkeys(self.cache['profiles']))
        self.write()

    def clear(self):
        # write empty cache file
        self.cache = self.default_key_values
        self.write()

    def clear_kubeconfig(self):
        # delete kube config if it exists
        kubeconfig = Path(self.base_path) / 'kube' / 'config'
        # kubeconfig.unlink(missing_ok=True)
        # removed for now, for 3.7 compatability
        try:
            kubeconfig.unlink()
        except FileNotFoundError:
            pass

    def get_credentials(self, profile_name: str, mode: str = 'awscredentialprocess'):
        try:
            ciphertext = self.cache[mode].get(profile_name.lower())
            if not ciphertext:
                return None
            return json.loads(self.string_encryptor.decrypt(ciphertext))
        except InvalidPassphraseException:  # if we cannot decrypt don't error - just make the API call to get the creds
            return None

    def save_credentials(self, profile_name: str, credentials: dict, mode: str = 'awscredentialprocess'):
        ciphertext = self.string_encryptor.encrypt(json.dumps(credentials, default=str))
        self.cache[mode][profile_name.lower()] = ciphertext
        self.write()

    def clear_credentials(self, profile_name: str, mode: str = 'awscredentialprocess'):
        self.cache[mode].pop(profile_name.lower(), None)
        self.write()

    @staticmethod
    def hash_banner(banner: dict) -> str:
        return hashlib.sha512(json.dumps(banner, default=str)).hexdigest()

    def banner_expired(self, tenant: str) -> bool:
        cached_banner_data = self.cache.get('banners', {}).get(tenant)
        expires = 0
        if cached_banner_data is not None:
            expires = cached_banner_data.get('expires', 0)
        return expires < int(time.time())

    def save_banner(self, tenant: str, banner: dict) -> bool:
        # if someone called this then we simply save the banner
        # regardless of whether the cached record has expired yet
        # as we assume the caller knows what they are doing

        cached_banner_data = self.cache.get('banners', {}).get(tenant, {})
        cached_hash = cached_banner_data.get('hash', '')
        new_hash = hashlib.sha512(json.dumps(banner, default=str, sort_keys=True).encode('utf-8')).hexdigest()
        self.cache['banners'][tenant] = {
            'hash': new_hash,
            'expires': int(time.time()) + (5 * 60)
        }
        self.write()

        # return True if the hashes have changed, False is they are equal
        return cached_hash != new_hash



import click
from ..helpers.build_britive import build_britive
from ..options.britive_options import britive_options
from ..helpers.profile_argument_decorator import click_smart_profile_argument


@click.group()
def request():
    """Provides functionality related to requesting approval to checkout a profile."""
    pass


@request.command()
@build_britive
@britive_options(names='justification,tenant,token,silent,passphrase,federation_provider')
@click_smart_profile_argument
def submit(ctx, justification, tenant, token, silent, passphrase, federation_provider, profile):
    """Submit a request to checkout a profile.

    Only applicable for profiles which require approval. This command will NOT block/wait until the request is
    approved or rejected. If you want to wait for the request to be approved, run the `checkout` command which will
    wait until the request is approved and then checkout the profile.

    This command takes 1 required argument `PROFILE`. This should be a string representation of the profile
    that should be checked out. Format is `application name/environment name/profile name`.
    """

    ctx.obj.britive.request_submit(
        profile=profile,
        justification=justification
    )


@request.command()
@build_britive
@britive_options(names='tenant,token,silent,passphrase,federation_provider')
@click_smart_profile_argument
def withdraw(ctx, tenant, token, silent, passphrase, federation_provider, profile):
    """Withdraw a request to checkout a profile.

    Only applicable for profiles which require approval.

    This command takes 1 required argument `PROFILE`. This should be a string representation of the profile
    that should be checked out. Format is `application name/environment name/profile name`.
    """

    ctx.obj.britive.request_withdraw(
        profile=profile
    )


@request.command()
@build_britive
@britive_options(names='tenant,token,silent,passphrase,federation_provider')
@click.argument('request-id')
def approve(ctx, tenant, token, silent, passphrase, federation_provider, request_id):
    """Approve a request to checkout a profile.

    This command takes 1 required argument `request-id`. Find the `request-id` via `ls approvals`.
    """

    ctx.obj.britive.request_disposition(
        request_id=request_id,
        decision='approve'
    )

@request.command()
@build_britive
@britive_options(names='tenant,token,silent,passphrase,federation_provider')
@click.argument('request-id')
def reject(ctx, tenant, token, silent, passphrase, federation_provider, request_id):
    """Reject a request to checkout a profile.

    This command takes 1 required argument `request-id`. Find the `request-id` via `ls approvals`.
    """

    ctx.obj.britive.request_disposition(
        request_id=request_id,
        decision='reject'
    )



def profile_split(profile):
    def str_escape_split(str_to_escape, delimiter=',', escape='\\'):
        if len(delimiter) > 1 or len(escape) > 1:
            raise ValueError("Either delimiter or escape must be an one char value")
        token = ''
        escaped = False
        for c in str_to_escape:
            if c == escape:
                if escaped:
                    token += escape
                    escaped = False
                else:
                    escaped = True
                continue
            if c == delimiter:
                if not escaped:
                    yield token
                    token = ''
                else:
                    token += c
                    escaped = False
            else:
                if escaped:
                    token += escape
                    escaped = False
                token += c
        yield token
    return list(str_escape_split(profile, delimiter='/', escape='\\'))



import click

option = click.option(
    '--federation-provider', '-P',
    help='Use a federation provider available in the Britive Python SDK for auto token creation. '
         'See CLI documentation at https://britive.github.io/python-cli/ for acceptable values.',
    default=None,
    show_default=True
)
